@@ -1,17 +1,17 @@
-/* 
+/*
    Copyright (C) Andrew Tridgell 1998,
-   Con Kolivas 2006-2008
-   
+   Con Kolivas 2006-2009
+
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
-   
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-   
+
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
@@ -19,81 +19,156 @@
 /* multiplex N streams into a file - the streams are passed
    through different compressors */
 
-#include "rzip.h"
+/* Need a definitition for FILE for old bzlib.h */
+#include <stdio.h>
 #include <bzlib.h>
 #include <zlib.h>
-
-/* since we have both zlib.h and lzmalib.h, omit lzmalib.h
- * and just but the two function definitions it needs.
- * The two headers are identical and cause compiler warnings.
-*/
-
-ZEXTERN int ZEXPORT lzma_compress OF((Bytef *dest,   uLongf *destLen,
-                                  const Bytef *source, uLong sourceLen,
-                                  int level, unsigned short lc,
-				  unsigned short lp, unsigned short pb,
-				  unsigned threads));
-ZEXTERN int ZEXPORT lzma_uncompress OF((Bytef *dest,   uLongf *destLen,
-                                   const Bytef *source, uLong sourceLen, int level,
-				   unsigned short lc, unsigned short lp, 
-				   unsigned short pb));
-
-#define CTYPE_NONE 3
-#define CTYPE_BZIP2 4
-#define CTYPE_LZO 5
-#define CTYPE_LZMA 6
-#define CTYPE_GZIP 7
+#include "rzip.h"
+/* LZMA C Wrapper */
+#include "lzma/C/LzmaLib.h"
 
 #define STREAM_BUFSIZE (1024 * 1024 * 10)
 
-typedef uint16 u16;
-typedef uint32 u32;
-
 struct stream {
-	u32 last_head;
+	i64 last_head;
 	uchar *buf;
-	unsigned int buflen;
-	unsigned int bufp;
+	i64 buflen;
+	i64 bufp;
 };
 
 struct stream_info {
 	struct stream *s;
 	int num_streams;
 	int fd;
-	u32 bufsize;
-	u32 cur_pos;
-	off_t initial_pos;
-	off_t total_read;
+	i64 bufsize;
+	i64 cur_pos;
+	i64 initial_pos;
+	i64 total_read;
 };
 
 /* just to keep things clean, declare function here
  * but move body to the end since it's a work function
 */
- 
-static int lzo_compresses( struct stream *s );
+
+static int lzo_compresses(struct stream *s);
+
+#ifndef HAS_MEMSTREAM
+#define fmemopen fake_fmemopen
+#define open_memstream fake_open_memstream
+#define memstream_update_buffer fake_open_memstream_update_buffer
+
+static FILE *fake_fmemopen(void *buf, size_t buflen, char *mode)
+{
+	FILE *in;
+
+	if (strcmp(mode, "r"))
+                fatal("fake_fmemopen only supports mode \"r\".");
+	in = tmpfile();
+	if (!in)
+	        return NULL;
+	if (fwrite(buf, buflen, 1, in) != 1)
+	        return NULL;
+	rewind(in);
+        return in;
+}
+
+static FILE *fake_open_memstream(char **buf, size_t *length)
+{
+	FILE *out;
+
+	if (buf == NULL || length == NULL)
+	        fatal("NULL parameter to fake_open_memstream");
+	out = tmpfile();
+	if (!out)
+	        return NULL;
+	return out;
+}
+
+static int fake_open_memstream_update_buffer(FILE *fp, uchar **buf, size_t *length)
+{
+	long original_pos = ftell(fp);
+
+	if (fseek(fp, 0, SEEK_END) != 0)
+		return -1;
+	*length = ftell(fp);
+	rewind(fp);
+	*buf = (uchar *)malloc(*length);
+	if (!*buf)
+	        return -1;
+	if (fread(*buf, *length, 1, fp) != 1)
+	        return -1;
+	if (fseek(fp, original_pos, SEEK_SET) != 0)
+	        return -1;
+	return 0;
+}
+#else
+static int memstream_update_buffer(FILE *fp, uchar **buf, size_t *length)
+{
+	return 0;
+}
+#endif
 
 /*
   ***** COMPRESSION FUNCTIONS *****
-  
-  BZIP, GZIP, LZMA, LZO
-  
+
+  ZPAQ, BZIP, GZIP, LZMA, LZO
+
   try to compress a buffer. If compression fails for whatever reason then
   leave uncompressed. Return the compression type in c_type and resulting
   length in c_len
 */
 
-static void bzip2_compress_buf(struct stream *s, int *c_type, u32 *c_len)
+static void zpaq_compress_buf(struct stream *s, int *c_type, i64 *c_len)
 {
 	uchar *c_buf;
-	unsigned int dlen = s->buflen-1;
+	FILE *in, *out;
+	size_t dlen;
+
+	if (!lzo_compresses(s))
+		return;
+
+	in = fmemopen(s->buf, s->buflen, "r");
+	if (!in)
+		fatal("Failed to fmemopen in zpaq_compress_buf\n");
+	out = open_memstream((char **)&c_buf, &dlen);
+	if (!out)
+		fatal("Failed to open_memstream in zpaq_compress_buf\n");
+
+	zpipe_compress(in, out, s->buflen, (int)(control.flags & FLAG_SHOW_PROGRESS));
+
+	if (memstream_update_buffer(out, &c_buf, &dlen) != 0)
+	        fatal("Failed to memstream_update_buffer in zpaq_compress_buf");
+
+	fclose(in);
+	fclose(out);
+
+	if ((i64)dlen >= *c_len) {
+		/* Incompressible, leave as CTYPE_NONE */
+		free(c_buf);
+		return;
+	}
+
+	*c_len = dlen;
+	free(s->buf);
+	s->buf = c_buf;
+	*c_type = CTYPE_ZPAQ;
+}
+
+static void bzip2_compress_buf(struct stream *s, int *c_type, i64 *c_len)
+{
+	uchar *c_buf;
+	u32 dlen = s->buflen;
+
+	if (!lzo_compresses(s))
+		return;
 
 	c_buf = malloc(dlen);
 	if (!c_buf)
 		return;
 
-	if (BZ2_bzBuffToBuffCompress((char*)c_buf, &dlen, (char*)s->buf, s->buflen, 
+	if (BZ2_bzBuffToBuffCompress((char*)c_buf, &dlen, (char*)s->buf, s->buflen,
 				     control.compression_level, 0,
-				     control.compression_level*10) != BZ_OK) {
+				     control.compression_level * 10) != BZ_OK) {
 		free(c_buf);
 		return;
 	}
@@ -110,22 +185,21 @@ static void bzip2_compress_buf(struct stream *s, int *c_type, u32 *c_len)
 	*c_type = CTYPE_BZIP2;
 }
 
-static void gzip_compress_buf(struct stream *s, int *c_type, u32 *c_len)
+static void gzip_compress_buf(struct stream *s, int *c_type, i64 *c_len)
 {
 	uchar *c_buf;
-	unsigned long dlen = s->buflen-1;
+	unsigned long dlen = s->buflen;
 
 	c_buf = malloc(dlen);
 	if (!c_buf)
 		return;
 
-	if (compress2(c_buf, &dlen, s->buf, s->buflen, 
-				     control.compression_level) != Z_OK) {
+	if (compress2(c_buf, &dlen, s->buf, s->buflen, control.compression_level) != Z_OK) {
 		free(c_buf);
 		return;
 	}
 
-	if (dlen >= *c_len) {
+	if ((i64)dlen >= *c_len) {
 		/* Incompressible, leave as CTYPE_NONE */
 		free(c_buf);
 		return;
@@ -137,30 +211,55 @@ static void gzip_compress_buf(struct stream *s, int *c_type, u32 *c_len)
 	*c_type = CTYPE_GZIP;
 }
 
-static void lzma_compress_buf(struct stream *s, int *c_type, u32 *c_len)
+static void lzma_compress_buf(struct stream *s, int *c_type, i64 *c_len)
 {
-	uchar *c_buf = NULL;
-	unsigned long dlen = s->buflen - 1;
+	uchar *c_buf;
+	size_t dlen;
+	size_t prop_size = 5; /* return value for lzma_properties */
+	int lzma_ret;
 
 	if (!lzo_compresses(s))
 		goto out;
 
+	dlen = s->buflen;
 	c_buf = malloc(dlen);
 	if (!c_buf)
 		return;
 
 	if (control.flags & FLAG_SHOW_PROGRESS) {
-		printf("\tProgress percentage pausing during lzma compression...");
-		fflush(stdout);
+		fprintf(stderr, "\tProgress percentage pausing during lzma compression...");
+		fflush(stderr);
 	}
-	if (lzma_compress(c_buf,&dlen,(uchar*)s->buf, s->buflen, control.compression_level, 
-			control.lc, control.lp, control.pb, control.threads) != 0) {
+	/* with LZMA SDK 4.63, we pass compression level and threads only
+	 * and receive properties in control->lzma_properties */
+
+	lzma_ret = LzmaCompress(c_buf, &dlen, s->buf, (size_t)s->buflen, control.lzma_properties, &prop_size, control.compression_level,
+			      0, /* dict size. set default */
+			      -1, -1, -1, -1, /* lc, lp, pb, fb */
+			      control.threads);
+	if (lzma_ret != SZ_OK) {
+		switch (lzma_ret) {
+			case SZ_ERROR_MEM:
+				err_msg("\nLZMA ERROR: %d. Try a smaller compression window.\n", SZ_ERROR_MEM);
+				break;
+			case SZ_ERROR_PARAM:
+				err_msg("\nLZMA Parameter ERROR: %d. This should not happen.\n", SZ_ERROR_PARAM);
+				break;
+			case SZ_ERROR_OUTPUT_EOF:
+				err_msg("\nHarmless LZMA Output Buffer Overflow error: %d. Incompressible block.\n", SZ_ERROR_OUTPUT_EOF);
+				break;
+			case SZ_ERROR_THREAD:
+				err_msg("\nLZMA Multi Thread ERROR: %d. This should not happen.\n", SZ_ERROR_THREAD);
+				break;
+			default:
+				err_msg("Unidentified LZMA ERROR: %d. This should not happen.\n", lzma_ret);
+				break;
+		}
 		/* can pass -1 if not compressible! Thanks Lasse Collin */
 		free(c_buf);
 		goto out;
 	}
-
-	if (dlen >= *c_len) {
+	if ((i64)dlen >= *c_len) {
 		/* Incompressible, leave as CTYPE_NONE */
 		free(c_buf);
 		goto out;
@@ -172,13 +271,13 @@ static void lzma_compress_buf(struct stream *s, int *c_type, u32 *c_len)
 	*c_type = CTYPE_LZMA;
 out:
 	if (control.flags & FLAG_VERBOSITY_MAX)
-		printf("\n");
-	else if ((control.flags & FLAG_SHOW_PROGRESS || control.flags & FLAG_VERBOSITY )) 
-		printf("\r\t                                                                                      \r");
-	fflush(stdout);
+		fprintf(stderr, "\n");
+	else if ((control.flags & FLAG_SHOW_PROGRESS || control.flags & FLAG_VERBOSITY ))
+		fprintf(stderr, "\r\t                                                                                      \r");
+	fflush(stderr);
 }
 
-static void lzo_compress_buf(struct stream *s, int *c_type, u32 *c_len)
+static void lzo_compress_buf(struct stream *s, int *c_type, i64 *c_len)
 {
 	uchar *c_buf;
 	lzo_bytep wrkmem;
@@ -194,7 +293,8 @@ static void lzo_compress_buf(struct stream *s, int *c_type, u32 *c_len)
 	if (!c_buf)
 		goto out_free;
 
-	return_var = lzo1x_1_compress((uchar *)s->buf,in_len,(uchar *)c_buf,&dlen,wrkmem);
+	return_var = lzo1x_1_compress((uchar *)s->buf, in_len, (uchar *)c_buf,
+				      &dlen,wrkmem);
 
 	if (dlen >= in_len){
 		/* Incompressible, leave as CTYPE_NONE */
@@ -212,16 +312,51 @@ out_free:
 
 /*
   ***** DECOMPRESSION FUNCTIONS *****
-  
-  BZIP, GZIP, LZMA, LZO
-  
+
+  ZPAQ, BZIP, GZIP, LZMA, LZO
+
   try to decompress a buffer. Return 0 on success and -1 on failure.
 */
 
-static int bzip2_decompress_buf(struct stream *s, u32 c_len)
+static int zpaq_decompress_buf(struct stream *s)
 {
 	uchar *c_buf;
-	unsigned int dlen = s->buflen;
+	FILE *in, *out;
+	size_t dlen;
+
+	in = fmemopen(s->buf, s->buflen, "r");
+	if (!in) {
+		err_msg("Failed to fmemopen in zpaq_decompress_buf\n");
+		return -1;
+	}
+	out = open_memstream((char **)&c_buf, &dlen);
+	if (!out) {
+		err_msg("Failed to open_memstream in zpaq_decompress_buf\n");
+		return -1;
+	}
+
+	zpipe_decompress(in, out, s->buflen, (int)(control.flags & FLAG_SHOW_PROGRESS));
+
+	if (memstream_update_buffer(out, &c_buf, &dlen) != 0)
+	        fatal("Failed to memstream_update_buffer in zpaq_decompress_buf");
+
+	fclose(in);
+	fclose(out);
+	free(s->buf);
+	s->buf = c_buf;
+
+	if ((i64)dlen != s->buflen) {
+		err_msg("Inconsistent length after decompression. Got %d bytes, expected %d\n", dlen, s->buflen);
+		return -1;
+	}
+
+	return 0;
+}
+
+static int bzip2_decompress_buf(struct stream *s, i64 c_len)
+{
+	uchar *c_buf;
+	u32 dlen = s->buflen;
 	int bzerr;
 
 	c_buf = s->buf;
@@ -246,7 +381,7 @@ static int bzip2_decompress_buf(struct stream *s, u32 c_len)
 	return 0;
 }
 
-static int gzip_decompress_buf(struct stream *s, u32 c_len)
+static int gzip_decompress_buf(struct stream *s, i64 c_len)
 {
 	uchar *c_buf;
 	unsigned long dlen = s->buflen;
@@ -265,7 +400,7 @@ static int gzip_decompress_buf(struct stream *s, u32 c_len)
 		return -1;
 	}
 
-	if (dlen != s->buflen) {
+	if ((i64)dlen != s->buflen) {
 		err_msg("Inconsistent length after decompression. Got %d bytes, expected %d\n", dlen, s->buflen);
 		return -1;
 	}
@@ -274,10 +409,10 @@ static int gzip_decompress_buf(struct stream *s, u32 c_len)
 	return 0;
 }
 
-static int lzma_decompress_buf(struct stream *s, u32 c_len)
+static int lzma_decompress_buf(struct stream *s, size_t c_len)
 {
 	uchar *c_buf;
-	unsigned long dlen = s->buflen;
+	size_t dlen = (size_t)s->buflen;
 	int lzmaerr;
 
 	c_buf = s->buf;
@@ -287,14 +422,15 @@ static int lzma_decompress_buf(struct stream *s, u32 c_len)
 		return -1;
 	}
 
-	lzmaerr = lzma_uncompress((uchar*)s->buf, &dlen, (uchar*)c_buf, c_len, control.compression_level,
-			control.lc, control.lp, control.pb);
+	/* With LZMA SDK 4.63 we pass control.lzma_properties
+	 * which is needed for proper uncompress */
+	lzmaerr = LzmaUncompress(s->buf, &dlen, c_buf, &c_len, control.lzma_properties, 5);
 	if (lzmaerr != 0) {
 		err_msg("Failed to decompress buffer - lzmaerr=%d\n", lzmaerr);
 		return -1;
 	}
 
-	if (dlen != s->buflen) {
+	if ((i64)dlen != s->buflen) {
 		err_msg("Inconsistent length after decompression. Got %d bytes, expected %d\n", dlen, s->buflen);
 		return -1;
 	}
@@ -303,7 +439,7 @@ static int lzma_decompress_buf(struct stream *s, u32 c_len)
 	return 0;
 }
 
-static int lzo_decompress_buf(struct stream *s, u32 c_len)
+static int lzo_decompress_buf(struct stream *s, i64 c_len)
 {
 	uchar *c_buf;
 	lzo_uint dlen = s->buflen;
@@ -322,7 +458,7 @@ static int lzo_decompress_buf(struct stream *s, u32 c_len)
 		return -1;
 	}
 
-	if (dlen != s->buflen) {
+	if ((i64)dlen != s->buflen) {
 		err_msg("Inconsistent length after decompression. Got %d bytes, expected %d\n", dlen, s->buflen);
 		return -1;
 	}
@@ -333,16 +469,67 @@ static int lzo_decompress_buf(struct stream *s, u32 c_len)
 
 /* WORK FUNCTIONS */
 
-/* write to a file, return 0 on success and -1 on failure */
-static int write_buf(int f, uchar *p, int len)
+const i64 one_g = 1000 * 1024 * 1024;
+
+/* This is a custom version of write() which writes in 1GB chunks to avoid
+   the overflows at the >= 2GB mark thanks to 32bit fuckage. This should help
+   even on the rare occasion write() fails to write 1GB as well. */
+ssize_t write_1g(int fd, void *buf, i64 len)
 {
-	int ret;
-	ret = write(f, p, len);
+	i64 total, offset;
+	ssize_t ret;
+	uchar *offset_buf = buf;
+
+	total = offset = 0;
+	while (len > 0) {
+		if (len > one_g)
+			ret = one_g;
+		else
+			ret = len;
+		ret = write(fd, offset_buf, (size_t)ret);
+		if (ret < 0)
+			return ret;
+		len -= ret;
+		offset_buf += ret;
+		total += ret;
+	}
+	return total;
+}
+
+/* Ditto for read */
+ssize_t read_1g(int fd, void *buf, i64 len)
+{
+	i64 total, offset;
+	ssize_t ret;
+	uchar *offset_buf = buf;
+
+	total = offset = 0;
+	while (len > 0) {
+		if (len > one_g)
+			ret = one_g;
+		else
+			ret = len;
+		ret = read(fd, offset_buf, (size_t)ret);
+		if (ret < 0)
+			return ret;
+		len -= ret;
+		offset_buf += ret;
+		total += ret;
+	}
+	return total;
+}
+
+/* write to a file, return 0 on success and -1 on failure */
+static int write_buf(int f, uchar *p, i64 len)
+{
+	ssize_t ret;
+
+	ret = write_1g(f, p, (size_t)len);
 	if (ret == -1) {
 		err_msg("Write of length %d failed - %s\n", len, strerror(errno));
 		return -1;
 	}
-	if (ret != len) {
+	if (ret != (ssize_t)len) {
 		err_msg("Partial write!? asked for %d bytes but got %d\n", len, ret);
 		return -1;
 	}
@@ -355,34 +542,25 @@ static int write_u8(int f, uchar v)
 	return write_buf(f, &v, 1);
 }
 
-/* write a short */
-static int write_u16(int f, u16 v)
+/* write a i64 */
+static int write_i64(int f, i64 v)
 {
-	uchar p[2];
-	p[0] = v&0xFF;
-	p[1] = (v>>8)&0xFF;
-	return write_buf(f, p, 2);
-}
-
-/* write a long */
-static int write_u32(int f, u32 v)
-{
-	if (write_u16(f, v&0xFFFF) != 0 ||
-	    write_u16(f, (v>>16)) != 0) {
+	if (write_buf(f, (uchar *)&v, 8))
 		return -1;
-	}
+
 	return 0;
 }
 
-static int read_buf(int f, uchar *p, int len)
+static int read_buf(int f, uchar *p, i64 len)
 {
-	int ret;
-	ret = read(f, p, len);
+	ssize_t ret;
+
+	ret = read_1g(f, p, (size_t)len);
 	if (ret == -1) {
 		err_msg("Read of length %d failed - %s\n", len, strerror(errno));
 		return -1;
 	}
-	if (ret != len) {
+	if (ret != (ssize_t)len) {
 		err_msg("Partial read!? asked for %d bytes but got %d\n", len, ret);
 		return -1;
 	}
@@ -394,36 +572,24 @@ static int read_u8(int f, uchar *v)
 	return read_buf(f, v, 1);
 }
 
-static int read_u16(int f, u16 *v)
+static int read_u32(int f, u32 *v)
 {
-	uchar p[2];
-	if (read_buf(f, p, 2) != 0) {
+	if (read_buf(f, (uchar *)v, 4))
 		return -1;
-	}
-	*v = (p[1]<<8) | p[0];
 	return 0;
 }
 
-static int read_u32(int f, u32 *v)
+static int read_i64(int f, i64 *v)
 {
-	u16 v1, v2;
-
-	if (read_u16(f, &v1) != 0) {
+	if (read_buf(f, (uchar *)v, 8))
 		return -1;
-	}
-	if (read_u16(f, &v2) != 0) {
-		return -1;
-	}
-	*v = v2;
-	*v <<= 16;
-	*v |= v1;
 	return 0;
 }
 
 /* seek to a position within a set of streams - return -1 on failure */
-static int seekto(struct stream_info *sinfo, u32 pos)
+static int seekto(struct stream_info *sinfo, i64 pos)
 {
-	off_t spos = pos + sinfo->initial_pos;
+	i64 spos = pos + sinfo->initial_pos;
 	if (lseek(sinfo->fd, spos, SEEK_SET) != spos) {
 		err_msg("Failed to seek to %d in stream\n", pos);
 		return -1;
@@ -433,27 +599,41 @@ static int seekto(struct stream_info *sinfo, u32 pos)
 
 /* open a set of output streams, compressing with the given
    compression level and algorithm */
-void *open_stream_out(int f, int n, unsigned int limit)
+void *open_stream_out(int f, int n, i64 limit)
 {
+	unsigned cwindow = control.window;
 	int i;
 	struct stream_info *sinfo;
 
 	sinfo = malloc(sizeof(*sinfo));
-	if (!sinfo) {
+	if (!sinfo)
 		return NULL;
-	}
 
 	sinfo->num_streams = n;
 	sinfo->cur_pos = 0;
 	sinfo->fd = f;
+
 	/* 10MB streams for non lzma compress. There is virtually no gain
-	   in lzo and bzip2 with larger streams. With lzma, however, the
-	   larger the buffer, the better the compression so we make it as
-	   large as the window */
-	if (control.flags & FLAG_NOT_LZMA)
-		sinfo->bufsize = STREAM_BUFSIZE;
+	   in lzo, gzip and bzip2 with larger streams. With lzma and zpaq,
+	   however, the larger the buffer, the better the compression so we
+	   make it as large as the window up to the limit the compressor
+	   will take */
+	if (LZMA_COMPRESS(control.flags)) {
+		if (sizeof(long) == 4) {
+			/* Largest window supported on lzma 32bit is 600MB */
+			if (cwindow > 6)
+				cwindow = 6;
+		}
+		/* Largest window supported on lzma 64bit is 4GB */
+		if (cwindow > 40)
+			cwindow = 40;
+	}
+
+	if (LZMA_COMPRESS(control.flags) || (control.flags & FLAG_ZPAQ_COMPRESS))
+		sinfo->bufsize = STREAM_BUFSIZE * 10 * cwindow;
 	else
-		sinfo->bufsize = STREAM_BUFSIZE * 10 * control.window;
+		sinfo->bufsize = STREAM_BUFSIZE;
+
 	/* No point making the stream larger than the amount of data */
 	if (limit && limit < sinfo->bufsize)
 		sinfo->bufsize = limit;
@@ -465,26 +645,29 @@ void *open_stream_out(int f, int n, unsigned int limit)
 		return NULL;
 	}
 
-	for (i=0;i<n;i++) {
+	for (i = 0; i < n; i++) {
 		sinfo->s[i].buf = malloc(sinfo->bufsize);
-		if (!sinfo->s[i].buf) goto failed;
+		if (!sinfo->s[i].buf)
+			goto failed;
 	}
 
 	/* write the initial headers */
-	for (i=0;i<n;i++) {
-		sinfo->s[i].last_head = sinfo->cur_pos + 9;
+	for (i = 0; i < n; i++) {
+		sinfo->s[i].last_head = sinfo->cur_pos + 17;
 		write_u8(sinfo->fd, CTYPE_NONE);
-		write_u32(sinfo->fd, 0);
-		write_u32(sinfo->fd, 0);
-		write_u32(sinfo->fd, 0);
-		sinfo->cur_pos += 13;
+		write_i64(sinfo->fd, 0);
+		write_i64(sinfo->fd, 0);
+		write_i64(sinfo->fd, 0);
+		sinfo->cur_pos += 25;
 	}
 	return (void *)sinfo;
 
 failed:
-	for (i=0;i<n;i++) {
-		if (sinfo->s[i].buf) free(sinfo->s[i].buf);
+	for (i = 0; i < n; i++) {
+		if (sinfo->s[i].buf)
+			free(sinfo->s[i].buf);
 	}
+
 	free(sinfo);
 	return NULL;
 }
@@ -492,13 +675,13 @@ failed:
 /* prepare a set of n streams for reading on file descriptor f */
 void *open_stream_in(int f, int n)
 {
+	i64 header_length;
 	int i;
 	struct stream_info *sinfo;
 
 	sinfo = calloc(sizeof(*sinfo), 1);
-	if (!sinfo) {
+	if (!sinfo)
 		return NULL;
-	}
 
 	sinfo->num_streams = n;
 	sinfo->fd = f;
@@ -510,43 +693,57 @@ void *open_stream_in(int f, int n)
 		return NULL;
 	}
 
-	for (i=0;i<n;i++) {
+	for (i = 0; i < n; i++) {
 		uchar c;
-		u32 v1, v2;
+		i64 v1, v2;
 
-	again:
-		if (read_u8(f, &c) != 0) {
-			goto failed;
-		}
-		if (read_u32(f, &v1) != 0) {
-			goto failed;
-		}
-		if (read_u32(f, &v2) != 0) {
-			goto failed;
-		}
-		if (read_u32(f, &sinfo->s[i].last_head) != 0) {
+again:
+		if (read_u8(f, &c) != 0)
 			goto failed;
+
+		/* Compatibility crap for versions < 0.40 */
+		if (control.major_version == 0 && control.minor_version < 4) {
+			u32 v132, v232, last_head32;
+
+			if (read_u32(f, &v132) != 0)
+				goto failed;
+			if (read_u32(f, &v232) != 0)
+				goto failed;
+			if (read_u32(f, &last_head32) != 0)
+				goto failed;
+
+			v1 = v132;
+			v2 = v232;
+			sinfo->s[i].last_head = last_head32;
+			header_length = 13;
+		} else {
+			if (read_i64(f, &v1) != 0)
+				goto failed;
+			if (read_i64(f, &v2) != 0)
+				goto failed;
+			if (read_i64(f, &sinfo->s[i].last_head) != 0)
+				goto failed;
+			header_length = 25;
 		}
 
-		if (c == CTYPE_NONE && v1==0 && v2==0 && sinfo->s[i].last_head==0 &&
-		    i == 0) {
+		if (c == CTYPE_NONE && v1 == 0 && v2 == 0 && sinfo->s[i].last_head == 0 && i == 0) {
 			err_msg("Enabling stream close workaround\n");
-			sinfo->initial_pos += 13;
+			sinfo->initial_pos += header_length;
 			goto again;
 		}
 
-		sinfo->total_read += 13;
+		sinfo->total_read += header_length;
 
 		if (c != CTYPE_NONE) {
 			err_msg("Unexpected initial tag %d in streams\n", c);
 			goto failed;
 		}
 		if (v1 != 0) {
-			err_msg("Unexpected initial c_len %d in streams %d\n", v1, v2);
+			err_msg("Unexpected initial c_len %lld in streams %lld\n", v1, v2);
 			goto failed;
 		}
 		if (v2 != 0) {
-			err_msg("Unexpected initial u_len %d in streams\n", v2);
+			err_msg("Unexpected initial u_len %lld in streams\n", v2);
 			goto failed;
 		}
 	}
@@ -563,19 +760,16 @@ failed:
 static int flush_buffer(struct stream_info *sinfo, int stream)
 {
 	int c_type = CTYPE_NONE;
-	u32 c_len = sinfo->s[stream].buflen;
-	
-	if (seekto(sinfo, sinfo->s[stream].last_head) != 0) {
-		return -1;
-	}
-	if (write_u32(sinfo->fd, sinfo->cur_pos) != 0) {
-		return -1;
-	}
+	i64 c_len = sinfo->s[stream].buflen;
 
-	sinfo->s[stream].last_head = sinfo->cur_pos + 9;
-	if (seekto(sinfo, sinfo->cur_pos) != 0) {
+	if (seekto(sinfo, sinfo->s[stream].last_head) != 0)
+		return -1;
+	if (write_i64(sinfo->fd, sinfo->cur_pos) != 0)
+		return -1;
+
+	sinfo->s[stream].last_head = sinfo->cur_pos + 17;
+	if (seekto(sinfo, sinfo->cur_pos) != 0)
 		return -1;
-	}
 
 	if (!(control.flags & FLAG_NO_COMPRESS)) {
 		if (LZMA_COMPRESS(control.flags))
@@ -586,30 +780,29 @@ static int flush_buffer(struct stream_info *sinfo, int stream)
 			bzip2_compress_buf(&sinfo->s[stream], &c_type, &c_len);
 		else if (control.flags & FLAG_ZLIB_COMPRESS)
 			gzip_compress_buf(&sinfo->s[stream], &c_type, &c_len);
+		else if (control.flags & FLAG_ZPAQ_COMPRESS)
+			zpaq_compress_buf(&sinfo->s[stream], &c_type, &c_len);
 		else fatal("Dunno wtf compression to use!\n");
 	}
-//	printf("c_type %d, c_len %u\n",c_type, c_len);
 
 	if (write_u8(sinfo->fd, c_type) != 0 ||
-	    write_u32(sinfo->fd, c_len) != 0 ||
-	    write_u32(sinfo->fd, sinfo->s[stream].buflen) != 0 ||
-	    write_u32(sinfo->fd, 0) != 0) {
+	    write_i64(sinfo->fd, c_len) != 0 ||
+	    write_i64(sinfo->fd, sinfo->s[stream].buflen) != 0 ||
+	    write_i64(sinfo->fd, 0) != 0) {
 		return -1;
 	}
-	sinfo->cur_pos += 13;
+	sinfo->cur_pos += 25;
 
-	if (write_buf(sinfo->fd, sinfo->s[stream].buf, c_len) != 0) {
+	if (write_buf(sinfo->fd, sinfo->s[stream].buf, c_len) != 0)
 		return -1;
-	}
 	sinfo->cur_pos += c_len;
 
 	sinfo->s[stream].buflen = 0;
 
 	free(sinfo->s[stream].buf);
 	sinfo->s[stream].buf = malloc(sinfo->bufsize);
-	if (!sinfo->s[stream].buf) {
+	if (!sinfo->s[stream].buf)
 		return -1;
-	}
 	return 0;
 }
 
@@ -617,37 +810,47 @@ static int flush_buffer(struct stream_info *sinfo, int stream)
 static int fill_buffer(struct stream_info *sinfo, int stream)
 {
 	uchar c_type;
-	u32 u_len, c_len;
+	i64 header_length;
+	i64 u_len, c_len;
 
-	if (seekto(sinfo, sinfo->s[stream].last_head) != 0) {
+	if (seekto(sinfo, sinfo->s[stream].last_head) != 0)
 		return -1;
+
+	if (read_u8(sinfo->fd, &c_type) != 0)
+		return -1;
+	/* Compatibility crap for versions < 0.4 */
+	if (control.major_version == 0 && control.minor_version < 4) {
+		u32 c_len32, u_len32, last_head32;
+
+		if (read_u32(sinfo->fd, &c_len32) != 0)
+			return -1;
+		if (read_u32(sinfo->fd, &u_len32) != 0)
+			return -1;
+		if (read_u32(sinfo->fd, &last_head32) != 0)
+			return -1;
+		c_len = c_len32;
+		u_len = u_len32;
+		sinfo->s[stream].last_head = last_head32;
+		header_length = 13;
+	} else {
+		if (read_i64(sinfo->fd, &c_len) != 0)
+			return -1;
+		if (read_i64(sinfo->fd, &u_len) != 0)
+			return -1;
+		if (read_i64(sinfo->fd, &sinfo->s[stream].last_head) != 0)
+			return -1;
+		header_length = 25;
 	}
 
-	if (read_u8(sinfo->fd, &c_type) != 0) {
-		return -1;
-	}
-	if (read_u32(sinfo->fd, &c_len) != 0) {
-		return -1;
-	}
-	if (read_u32(sinfo->fd, &u_len) != 0) {
-		return -1;
-	}
-	if (read_u32(sinfo->fd, &sinfo->s[stream].last_head) != 0) {
-		return -1;
-	}
+	sinfo->total_read += header_length;
 
-	sinfo->total_read += 13;
-
-	if (sinfo->s[stream].buf) {
+	if (sinfo->s[stream].buf)
 		free(sinfo->s[stream].buf);
-	}
 	sinfo->s[stream].buf = malloc(u_len);
-	if (!sinfo->s[stream].buf) {
+	if (!sinfo->s[stream].buf)
 		return -1;
-	}
-	if (read_buf(sinfo->fd, sinfo->s[stream].buf, c_len) != 0) {
+	if (read_buf(sinfo->fd, sinfo->s[stream].buf, c_len) != 0)
 		return -1;
-	}
 
 	sinfo->total_read += c_len;
 
@@ -656,7 +859,7 @@ static int fill_buffer(struct stream_info *sinfo, int stream)
 
 	if (c_type != CTYPE_NONE) {
 		if (c_type == CTYPE_LZMA) {
-			if (lzma_decompress_buf(&sinfo->s[stream], c_len))
+			if (lzma_decompress_buf(&sinfo->s[stream], (size_t)c_len))
 				return -1;
 		} else if (c_type == CTYPE_LZO) {
 			if (lzo_decompress_buf(&sinfo->s[stream], c_len))
@@ -667,6 +870,9 @@ static int fill_buffer(struct stream_info *sinfo, int stream)
 		} else if (c_type == CTYPE_GZIP) {
 			if (gzip_decompress_buf(&sinfo->s[stream], c_len))
 				return -1;
+		} else if (c_type == CTYPE_ZPAQ) {
+			if (zpaq_decompress_buf(&sinfo->s[stream]))
+				return -1;
 		} else fatal("Dunno wtf decompression type to use!\n");
 	}
 
@@ -674,12 +880,14 @@ static int fill_buffer(struct stream_info *sinfo, int stream)
 }
 
 /* write some data to a stream. Return -1 on failure */
-int write_stream(void *ss, int stream, uchar *p, unsigned int len)
+int write_stream(void *ss, int stream, uchar *p, i64 len)
 {
 	struct stream_info *sinfo = ss;
 
 	while (len) {
-		int n = MIN(sinfo->bufsize - sinfo->s[stream].buflen, len);
+		i64 n;
+
+		n = MIN(sinfo->bufsize - sinfo->s[stream].buflen, len);
 
 		memcpy(sinfo->s[stream].buf+sinfo->s[stream].buflen, p, n);
 		sinfo->s[stream].buflen += n;
@@ -687,9 +895,8 @@ int write_stream(void *ss, int stream, uchar *p, unsigned int len)
 		len -= n;
 
 		if (sinfo->s[stream].buflen == sinfo->bufsize) {
-			if (flush_buffer(sinfo, stream) != 0) {
+			if (flush_buffer(sinfo, stream) != 0)
 				return -1;
-			}
 		}
 	}
 	return 0;
@@ -697,13 +904,15 @@ int write_stream(void *ss, int stream, uchar *p, unsigned int len)
 
 /* read some data from a stream. Return number of bytes read, or -1
    on failure */
-int read_stream(void *ss, int stream, uchar *p, unsigned int len)
+i64 read_stream(void *ss, int stream, uchar *p, i64 len)
 {
 	struct stream_info *sinfo = ss;
-	int ret=0;
+	i64 ret = 0;
 
 	while (len) {
-		int n = MIN(sinfo->s[stream].buflen-sinfo->s[stream].bufp, len);
+		i64 n;
+
+		n = MIN(sinfo->s[stream].buflen-sinfo->s[stream].bufp, len);
 
 		if (n > 0) {
 			memcpy(p, sinfo->s[stream].buf+sinfo->s[stream].bufp, n);
@@ -713,12 +922,11 @@ int read_stream(void *ss, int stream, uchar *p, unsigned int len)
 			ret += n;
 		}
 
-		if (len &&
-		    sinfo->s[stream].bufp == sinfo->s[stream].buflen) {
-			if (fill_buffer(sinfo, stream) != 0) {
+		if (len && sinfo->s[stream].bufp == sinfo->s[stream].buflen) {
+			if (fill_buffer(sinfo, stream) != 0)
 				return -1;
-			}
-			if (sinfo->s[stream].bufp == sinfo->s[stream].buflen) break;
+			if (sinfo->s[stream].bufp == sinfo->s[stream].buflen)
+				break;
 		}
 	}
 
@@ -730,12 +938,19 @@ int close_stream_out(void *ss)
 {
 	struct stream_info *sinfo = ss;
 	int i;
-	for (i=0;i<sinfo->num_streams;i++) {
-		if (sinfo->s[i].buflen != 0 &&
-		    flush_buffer(sinfo, i) != 0) {
-			return -1;
+
+	/* reallocate buffers to try and save space */
+	for (i = 0; i < sinfo->num_streams; i++) {
+		if (sinfo->s[i].buflen != 0) {
+			if (!realloc(sinfo->s[i].buf, sinfo->s[i].buflen))
+				fatal("Error Reallocating Output Buffer %d\n", i);
 		}
-		if (sinfo->s[i].buf) free(sinfo->s[i].buf);
+	}
+	for (i = 0; i < sinfo->num_streams; i++) {
+		if (sinfo->s[i].buflen != 0 && flush_buffer(sinfo, i) != 0)
+			return -1;
+		if (sinfo->s[i].buf)
+			free(sinfo->s[i].buf);
 	}
 	free(sinfo->s);
 	free(sinfo);
@@ -748,12 +963,12 @@ int close_stream_in(void *ss)
 	struct stream_info *sinfo = ss;
 	int i;
 
-	if (lseek(sinfo->fd, sinfo->initial_pos + sinfo->total_read, 
-		  SEEK_SET) != sinfo->initial_pos + sinfo->total_read) {
-		return -1;
-	}
-	for (i=0;i<sinfo->num_streams;i++) {
-		if (sinfo->s[i].buf) free(sinfo->s[i].buf);
+	if (lseek(sinfo->fd, sinfo->initial_pos + sinfo->total_read,
+		  SEEK_SET) != sinfo->initial_pos + sinfo->total_read)
+			return -1;
+	for (i = 0; i < sinfo->num_streams; i++) {
+		if (sinfo->s[i].buf)
+			free(sinfo->s[i].buf);
 	}
 
 	free(sinfo->s);
@@ -761,10 +976,10 @@ int close_stream_in(void *ss)
 	return 0;
 }
 
-/* As lzma is very slow and lzo very fast, it is worth doing a quick lzo pass
+/* As others are slow and lzo very fast, it is worth doing a quick lzo pass
    to see if there is any compression at all with lzo first. It is unlikely
-   that lzma will be able to compress if lzo is unable to drop a single byte
-   so do not compress any block with lzma that is incompressible by lzo. */
+   that others will be able to compress if lzo is unable to drop a single byte
+   so do not compress any block that is incompressible by lzo. */
 static int lzo_compresses(struct stream *s)
 {
 	lzo_bytep wrkmem=NULL;
@@ -773,86 +988,63 @@ static int lzo_compresses(struct stream *s)
 	lzo_int return_var;	/* lzo1x_1_compress does not return anything but LZO_OK */
 	uchar *c_buf = NULL, *test_buf = s->buf;
 	/* set minimum buffer test size based on the length of the test stream */
-	unsigned long buftest_size = (test_len > 5*STREAM_BUFSIZE ? STREAM_BUFSIZE : STREAM_BUFSIZE / 4096);
+	unsigned long buftest_size = (test_len > 5 * STREAM_BUFSIZE ? STREAM_BUFSIZE : STREAM_BUFSIZE / 4096);
 	int ret = 0;
 	int workcounter = 0;	/* count # of passes */
 	lzo_uint best_dlen = UINT_MAX; /* save best compression estimate */
 
-#if 0
-	/* This assumption is wrong */
-
-	/* if buffer length > window size * threshold, abort right away */
-	if ( (double) save_len > (double) ( control.window * 104857600 * control.threshold) ) {
-		if ( (control.flags & FLAG_VERBOSITY) || (control.flags & FLAG_VERBOSITY_MAX) ) {
-			printf("\tSkipping lzo compression test. Data chunk %ld > Window * Threshold. %5.2f%%",
-				save_len, 100 * 
-				((double) save_len / (double) (control.window * 104857600 * control.threshold)));
-			fflush(stdout);
-		}
-		ret=0;
-		goto out;
-	}
-#endif
+	if (control.threshold > 1)
+		return 1;
 	wrkmem = (lzo_bytep) malloc(LZO1X_1_MEM_COMPRESS);
 	if (wrkmem == NULL)
-		goto out;
+		fatal("Unable to allocate wrkmem in lzo_compresses\n");
 
 	in_len = MIN(test_len, buftest_size);
 	dlen = STREAM_BUFSIZE + STREAM_BUFSIZE / 16 + 64 + 3;
 
 	c_buf = malloc(dlen);
 	if (!c_buf)
-		goto out;
+		fatal("Unable to allocate c_buf in lzo_compresses\n");
 
 	if (control.flags & FLAG_SHOW_PROGRESS) {
-		printf("\tlzo testing for incompressible data...");
-		fflush(stdout);
+		fprintf(stderr, "\tlzo testing for incompressible data...");
+		fflush(stderr);
 	}
 
 	/* Test progressively larger blocks at a time and as soon as anything
 	   compressible is found, jump out as a success */
 	while (test_len > 0) {
 		workcounter++;
-		return_var = lzo1x_1_compress(test_buf,in_len,(uchar *)c_buf,&dlen,wrkmem);
+		return_var = lzo1x_1_compress(test_buf, in_len, (uchar *)c_buf, &dlen, wrkmem);
 
-		/* since this test can return false positives
-		 * i.e. a compressible part followed by a long random
-		 * stream of data, a threshold of 90% (-T 2) for lzo compression
-		 * is set by default, but can be changed at startup with the -T option.
-		 * The theory is that if lzo can't compress well, neither will lzma.
-		*/
-		
-		if ( dlen < best_dlen )
+		if (dlen < best_dlen)
 			best_dlen = dlen;	/* save best value */
 
-		if ( (double) dlen < (double) in_len * control.threshold ) {
+		if ((double) dlen < (double)in_len * control.threshold) {
 			ret = 1;
 			break;
 		}
 		/* expand and move buffer */
 		test_len -= in_len;
 		if (test_len) {
-			test_buf += (ptrdiff_t) in_len;
+			test_buf += (ptrdiff_t)in_len;
 			if (buftest_size < STREAM_BUFSIZE)
 				buftest_size <<= 1;
 			in_len = MIN(test_len, buftest_size);
 		}
 	}
-	if ( control.flags & FLAG_VERBOSITY_MAX )
-		printf("%s for chunk %ld. Compressed size = %5.2F%% of chunk, %d Passes\n", 
-			(ret==0?"FAILED - below threshold":"OK"), save_len, 
+	if (control.flags & FLAG_VERBOSITY_MAX)
+		fprintf(stderr, "%s for chunk %ld. Compressed size = %5.2F%% of chunk, %d Passes\n",
+			(ret == 0? "FAILED - below threshold" : "OK"), save_len,
 			100 * ((double) best_dlen / (double) in_len), workcounter);
-	else if ( control.flags & FLAG_VERBOSITY )
-		printf("%s\r", (ret==0?"FAILED - below threshold":"OK"));
+	else if (control.flags & FLAG_VERBOSITY)
+		fprintf(stderr, "%s\r", (ret == 0? "FAILED - below threshold" : "OK"));
 	else if (control.flags & FLAG_SHOW_PROGRESS)
-		printf("\r\t                                                      \r");
-	fflush(stdout);
+		fprintf(stderr, "\r\t                                                      \r");
+	fflush(stderr);
 
-out:
-	if (wrkmem)
-		free(wrkmem);
-	if (c_buf)
-		free(c_buf);
+	free(wrkmem);
+	free(c_buf);
 
 	return ret;
 }
